package main

import (
	"fmt"
	"sync"
	"time"
)

type LogSeverityType string

const (
	Info    LogSeverityType = "Info"
	Warning LogSeverityType = "Warning"
	Error   LogSeverityType = "Error"
	None    LogSeverityType = "None"
)

type LogMessage struct { // структура сообщений логера
	Severity LogSeverityType
	Message  string
}

type ActiveLogger struct { // данные Active Object
	logQueue  chan *LogMessage // канал логирования
	stopEvent chan struct{}    // канал завершения
	tmStart   time.Time        // время старта
	wg        sync.WaitGroup
}

func createActiveLogger() *ActiveLogger {
	return &ActiveLogger{
		logQueue:  make(chan *LogMessage, 10),
		stopEvent: make(chan struct{}),
		tmStart:   time.Now(),
	}
}

func (l *ActiveLogger) Log(message *LogMessage) {
	l.logQueue <- message
}

func (l *ActiveLogger) StartLogProcessor() {
	l.wg.Add(1)
	go func() { // loop Active Object
		defer l.wg.Done()
		for { // цикл обслуживания
			select {
			case message := <-l.logQueue:
				l.processMessage(message)
			case <-l.stopEvent:
				return
			}
		}
	}()
}

// фактическое логирование: тут может быть любая длинная работа...
func (l *ActiveLogger) processMessage(m *LogMessage) {
	out := func(m string) {
		dur := func() time.Duration {
			return time.Now().Sub(l.tmStart)
		}
		fmt.Printf("% 14v : %s", dur(), m)
	}
	l.wg.Add(1)
	go func() {
		defer l.wg.Done()
		out(fmt.Sprintf("логируется... [%s]:\t%s\n", m.Severity, m.Message))
		time.Sleep(500 * time.Millisecond)
		out(fmt.Sprintf("залогировано! [%s]:\t%s\n", m.Severity, m.Message))
	}()
}

func (l *ActiveLogger) StopLogProcessor() {
	message := &LogMessage{
		Message:  fmt.Sprintf("завершение логера"),
		Severity: None,
	}
	l.Log(message)     // сообщение завершения логера
	close(l.stopEvent) // закрыть канал завершения
	l.wg.Wait()        // ... и ожидать завершения рабочего цикла
}

func main() {
	logger := createActiveLogger() // создание Active Object
	logger.StartLogProcessor()     // запуск активности ...
	const N = 5
	for i := range N {
		var Severity LogSeverityType
		if i < N-1 {
			Severity = Info
		} else {
			Severity = Error
		}
		message := &LogMessage{
			Message:  fmt.Sprintf("сообщение номер %d", i),
			Severity: Severity,
		}
		logger.Log(message) // сообщение логеру ...
	}
	time.Sleep(2 * time.Second)
	logger.StopLogProcessor() // завершение логера
}
